import dataclasses
import json
from typing import Any, Optional

import pydantic

import pyckish.types.http_codes as status


@dataclasses.dataclass
class HTTPResponse:
    body: Any = dataclasses.field(default_factory=lambda: {})
    headers: Optional[dict] = dataclasses.field(default_factory=lambda: {})
    status_code: Optional[int] = None

    def __call__(self, response_config: dict) -> dict:
        response_dict = {'isBase64Encoded': False}
        if self.body:
            if isinstance(self.body, pydantic.BaseModel):
                response_dict['body'] = self.body.json(**response_config)
            else:
                response_dict['body'] = json.dumps(self.body)
        if self.headers:
            if 'Content-Type' not in self.headers.keys():
                self.headers['Content-Type'] = 'application/json'
        else:
            self.headers = {'Content-Type': 'application/json'}
        response_dict['headers'] = self.headers
        response_dict['statusCode'] = self.status_code if self.status_code else status.HTTP_201_CREATED
        return response_dict



import inspect
from typing import Callable, Any, Type

import pydantic

from pyckish import LambdaInputElement
from pyckish.exceptions.validation_error import ValidationError


class ParameterHandler:

    def __init__(self) -> None:
        self.__raw_parameters = {}
        self.__model_structure = {}

    def generate_parameter_value_dict_for_lambda_function(
            self,
            event: dict,
            context: dict,
            lambda_handler_function: Callable[..., Any]
    ) -> dict:
        func_parameters = inspect.signature(lambda_handler_function).parameters
        for parameter in func_parameters.values():
            annotation = self.__get_annotation(parameter)
            raw_argument = self.__extract_raw_argument_from_inputs(
                event, context, parameter
            )
            self.__raw_parameters[parameter.name] = raw_argument
            self.__model_structure[parameter.name] = (annotation, ...)
        model = self.__generate_model()
        return model.__dict__

    def __generate_model(self) -> pydantic.BaseModel:
        FunctionParameters = pydantic.create_model("FunctionParameters", **self.__model_structure)
        try:
            return FunctionParameters(**self.__raw_parameters)
        except pydantic.ValidationError as exc:
            raise ValidationError('Validation Error', detail=exc.errors())

    @staticmethod
    def __get_annotation(parameter: inspect.Parameter) -> Type:
        if parameter.annotation == inspect.Parameter.empty:
            return Any
        return parameter.annotation

    @staticmethod
    def __extract_raw_argument_from_inputs(
            event: dict,
            context: dict,
            parameter: inspect.Parameter
    ) -> Any:
        if not issubclass(type(parameter.default), LambdaInputElement):
            raw_argument = event.get(parameter.name, parameter.default)
            if parameter.default == inspect.Parameter.empty and raw_argument is None:
                raise ValidationError(f'Parameter {parameter.name} is missing')
            return raw_argument
        lambda_input_element = parameter.default
        lambda_input_element.parameter_name = parameter.name
        raw_argument = lambda_input_element.extract(event, context)
        return raw_argument



import inspect
import re
from abc import ABC, abstractmethod
from typing import Optional, Any

EMPTY = inspect.Parameter.empty


class LambdaInputElement(ABC):
    """
    Your custom Event Element needs to be a child of this class
    """

    def __init__(
            self,
            alias: Optional[str] = None,
            default: Any = EMPTY(),
            regex: Optional[str] = None
    ) -> None:
        self.alias = alias
        self.regex = regex
        self.default = default
        self.parameter_name: Optional[str] = None

    @abstractmethod
    def extract(self, event: dict, context: dict) -> Any:
        ...

    def select_key_for_extraction(self, keys: set[str]) -> str:
        if self.alias:
            return self.alias
        if self.regex:
            try:
                for key in keys:
                    if re.match(self.regex, key):
                        return key
            except re.error:
                raise ValueError(f'Bad regex pattern on pyckish: "{self.regex}"')
        return self.parameter_name



from pyckish import LambdaInputElement
from pyckish.exceptions.validation_error import ValidationError


class Method(LambdaInputElement):
    """
    Use this class on your Lambda Handler Function.
    Extracts the HTTP Method
    """

    def extract(self, event: dict, context: dict) -> str:
        try:
            return event['method']
        except KeyError:
            ValidationError('Method not present in the event')



from typing import Optional, Union


class ValidationError(Exception):

    def __init__(self, message: str, detail: Optional[Union[dict, list]] = None) -> None:
        super().__init__(message)
        self.__detail = detail if detail is not None else {}

    def detail(self) -> Union[dict, list]:
        return self.__detail



from pyckish import LambdaInputElement


class Event(LambdaInputElement):
    """
    Use this class on your Lambda Handler Function.
    It just returns the raw event.
    """

    def extract(self, event: dict, context: dict) -> dict:
        return event



from typing import Optional, Any

from pyckish import LambdaInputElement, EMPTY
from pyckish.exceptions.validation_error import ValidationError


class EventElement(LambdaInputElement):
    """
    Use this class on your Lambda Handler Function.
    Extracts a single key from the AWS Event
    """

    def __init__(
            self,
            alias: Optional[str] = None,
            default: Any = EMPTY(),
            regex: Optional[str] = None
    ) -> None:
        super().__init__(alias=alias, default=default, regex=regex)

    def extract(self, event: dict, context: dict) -> Any:
        key = self.select_key_for_extraction(set(event.keys()))
        try:
            return event[key]
        except (KeyError, AttributeError):
            if type(self.default) == EMPTY:
                raise ValidationError(f'The event parameter "{key}" could not be found')
            return self.default



from typing import Callable, Any, Optional

InboundInterceptor = Callable[[dict, dict], tuple[dict, dict]]
OutboundInterceptor = Callable[[Any], Any]


class InterceptorHandler:

    def __init__(
            self,
            inbound_interceptors: Optional[list[InboundInterceptor]],
            outbound_interceptors: Optional[list[OutboundInterceptor]]
    ) -> None:
        self.__inbound_interceptors = [] if inbound_interceptors is None else inbound_interceptors
        self.__outbound_interceptors = [] if outbound_interceptors is None else outbound_interceptors

    def execute_chain_of_inbound_interceptors(self, event: dict, context: dict) -> tuple[dict, dict]:
        for interceptor in self.__inbound_interceptors:
            event, context = interceptor(event, context)
        return event, context

    def execute_chain_of_outbound_interceptors(self, output: Any) -> tuple[dict, dict]:
        for interceptor in self.__outbound_interceptors:
            output = interceptor(output)
        return output



from typing import Callable, Any, Optional, Type

ExcHandler = Callable[[dict, dict, Exception], Any]


class ExceptionHandler:

    def __init__(self, exception_to_handler_mapping: Optional[dict[Type[Exception], ExcHandler]]) -> None:
        self.__exception_handling_dict = {} if exception_to_handler_mapping is None else exception_to_handler_mapping

    def deal_with_exception(
            self,
            event: dict,
            context: dict,
            exception: Exception
    ) -> Any:
        for t_exc in self.__exception_handling_dict.keys():
            if t_exc == Exception:
                continue
            if self.__exception_handling_dict.get(type(exception), None):
                result = self.__exception_handling_dict[type(exception)](event, context, exception)
                return result
            if isinstance(exception, t_exc):
                result = self.__exception_handling_dict[t_exc](event, context, exception)
                return result
        if self.__exception_handling_dict.get(Exception, None):
            result = self.__exception_handling_dict[Exception](event, context, exception)
            return result
        raise exception



import json
from typing import Any

from pyckish.lambda_input_element import LambdaInputElement


class Body(LambdaInputElement):
    """
    Use this class on your Lambda Handler Function.
    Extracts an HTTP Body
    """

    def __init__(self) -> None:
        super().__init__()

    def extract(self, event: dict, context: dict) -> Any:
        return json.loads(event.get('body', "{}"))



import functools
from typing import Callable, Any, Type, Optional

from pyckish.internal_handlers.exception_handler import ExceptionHandler, ExcHandler
from pyckish.internal_handlers.interceptor_handler import InterceptorHandler, InboundInterceptor, OutboundInterceptor
from pyckish.internal_handlers.parameter_handler import ParameterHandler
from pyckish.internal_handlers.response_handler import ResponseHandler, DictOrSetIntStr


class Lambda:
    """
    To use it just place the decorator @pyckish.Lambda() above your handler of your AWS Lambda function.
    Pyckish will extract, parse and validate pre-defined structures that resides in the event.

    For instance, when AWS API Gateway activates an AWS Lambda, data that were within an HTTP request will reside
    now in the Event. To quickly extract and parse that data, just like a modern back-end framework would do, use
    pyckish.

    """

    def __init__(
            self,
            is_http: bool = False,
            response_status_code: Optional[int] = None,
            response_headers: Optional[dict[str, Any]] = None,
            response_model_exclude_unset: bool = False,
            response_model_exclude_defaults: bool = False,
            response_model_exclude_none: bool = False,
            response_model_by_alias: bool = False,
            response_model_include: Optional[DictOrSetIntStr] = None,
            response_model_exclude: Optional[DictOrSetIntStr] = None,
            exception_to_handler_mapping: Optional[dict[Type[Exception], ExcHandler]] = None,
            inbound_interceptors: Optional[list[InboundInterceptor]] = None,
            outbound_interceptors: Optional[list[OutboundInterceptor]] = None
    ) -> None:
        self.__interceptor_handler = InterceptorHandler(inbound_interceptors, outbound_interceptors)
        self.__parameter_handler = ParameterHandler()
        self.__exception_handler = ExceptionHandler(exception_to_handler_mapping)
        self.__response_handler = ResponseHandler(
            is_http=is_http,
            headers=response_headers,
            status_code=response_status_code,
            exclude_unset=response_model_exclude_unset,
            exclude_defaults=response_model_exclude_defaults,
            exclude_none=response_model_exclude_none,
            by_alias=response_model_by_alias,
            include=response_model_include,
            exclude=response_model_exclude
        )

    def __call__(self, lambda_handler_function: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(lambda_handler_function)
        def wrapper(event: dict, context: dict) -> Any:
            try:
                intercepted_event, intercepted_context = \
                    self.__interceptor_handler.execute_chain_of_inbound_interceptors(event, context)

                parameter_value_dict = self.__parameter_handler.generate_parameter_value_dict_for_lambda_function(
                    intercepted_event, intercepted_context, lambda_handler_function
                )
                result = lambda_handler_function(**parameter_value_dict)
                result = self.__interceptor_handler.execute_chain_of_outbound_interceptors(result)
            except Exception as exception:
                result = self.__exception_handler.deal_with_exception(event, context, exception)
            return self.__response_handler.prepare_response(result)

        return wrapper
